#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Name: 
    
Purpose: wrappers around Fast Marching Method calculations of 
geodesic distance and travel times
    
Description:
    
Comments:

TODO:
    
@author: dmccormick
Author: David S. McCormick
Created on 2020.01.13
"""
#%% import and function definitions
### import modules
import numpy as np

import matplotlib.pyplot as plt

import skfmm

### constants
EMPTY_STRING = ''

def calculate_distances(layer: np.array, \
                       origin: tuple, \
                       zero_threshold: float=0) -> np.array:
    """
    Calculate geodesic distances using the Fast Marching Method
    
    NB: currently 2D only
    
    TODO: generalize to 3D
    TODO: test zero_threshold >= 0

    Parameters
    ----------
    layer : np.array
        2d array; can be binary or continous values; should be all non-negative.
    origin : tuple
        (row, col) location of zero point from which to calcualte geodesic distances.
    zero_threshold : float, optional
        value below which to set layer values to zero. 
        The default is None, so that only current zero values will be masked out.

    Returns
    -------
    distances : np.array or None
        if value at origin point is >= greater than the threshold, 
        return an np.array of same size as layer with geodesic distances from origin point.
        if origin point is at or below the threshold value, return None
    """
    assert(zero_threshold >= 0)

    ### geodesic distances from fast marching method with mask from zero or below a threshold value
    layer_masked = np.ma.masked_where(layer <= zero_threshold, layer)
    
    if layer_masked[origin] is np.ma.masked:
        # the origin point is masked and you can't calculate a valid distance
        # TODO: should there be a message or exception at this point?        
        distances_masked = None
    else:
        original_value = layer_masked[origin]
        layer_masked[origin] = 0.0 # zero contour for FMM calculation
        distances_masked = skfmm.distance(layer_masked)
        layer_masked[origin] = original_value # reset value for next calcuation; necessary?

    return distances_masked


def calculate_travel_times(layer: np.array, \
                          speed: np.array, \
                          origin: tuple, \
                          zero_threshold: float=0) -> np.array:
    """
    Calculate travel times using the Fast Marching Method
    
    NB: currently 2D only
    
    TODO: generalize to 3D
    TODO: deal with case of origin point is below threshold so is masked, so
    calculating a travel time is invalid
    TODO: insure that speeds are all non-negative and hopefully positive

    Parameters
    ----------
    layer : np.array
        2d array; can be binary or continous values; should be all non-negative.
    speed : np.array
        2d array of same shape as layer; 
        continous values only; should be all non-negative.
    origin : tuple
        (row, col) location of zero point from which to calcualte geodesic distances.
    zero_threshold : float, optional
        value below which to set layer values to zero. 
        The default is None, so that only current zero values will be masked out.

    Returns
    -------
    travel_times : np.array
        if value at origin point is >= greater than the threshold, 
        return an np.array of same size as layer with travel times from origin point.
        if origin point is at or below the threshold value, return None
    """
    assert(layer.shape == speed.shape), "layer and speed grids not the same shape"
    assert(zero_threshold >= 0), "zero_threshold < 0"
    
    layer_masked = np.ma.masked_where(layer <= zero_threshold, layer)
    
    if layer_masked[origin] is np.ma.masked:
        # the origin point is masked and you can't calculate a valid travel time
        # TODO: should there be a message or exception at this point?
        travel_times = None
    else:
        original_value = layer_masked[origin]
        layer_masked[origin] = 0 # origin point for the FMM calculation
        try:
            travel_times = skfmm.travel_time(phi=layer_masked, speed=layer, order=2) # this commonly fails in skfmm due to numerics issues
        except:
            travel_times = skfmm.travel_time(phi=layer_masked, speed=layer, order=1) # only works with 1st order; bug
            print("fast_marching_method:calculate_travel_times(): could not order=2; falling back to order=1")

        layer_masked[origin] = original_value # reset value for next calcuation; necessary?
    
    return travel_times


# %% testing
def main() :
    ### simple grid
    from scipy import stats

    nrows = 60
    ncols = 200
    
    phi_min = 0.1
    phi_max = 0.3
    phi_diff = phi_max - phi_min
    
    x = np.linspace(0, 3 * np.pi, ncols)
    y = np.linspace(0, 2 * np.pi, nrows)
    phi_x = np.sin(x)
    phi_y = np.sin(y)

    grid = np.ones(shape=(nrows, ncols), dtype=float)
    
    for row in range(nrows):
        for col in range(ncols):
            grid[row, col] = phi_x[col] + phi_y[row]
    
    grid_min = grid.min()
    grid_max = grid.max()
    grid_range = grid_max - grid_min
    grid = (grid - grid_min) / grid_range * phi_diff + phi_min
    
    plt.contour(grid, levels=30)
    plt.colorbar()
    plt.title("grid values for phi")
    plt.show()
    
    print("grid stats:", stats.describe(grid.flatten()))
    print()
    
    ### compute distances
    # set zero point for FMM zero contour
    zero_point = (2, 2)
    
    
    distances = calculate_distances(grid, zero_point)
    
    plt.contour(distances, levels=50)
    plt.colorbar()
    plt.title(f'distances from {zero_point}')
    plt.show()
    
    plt.hist(distances.flatten(), bins=50)
    plt.title("distances")
    plt.show()
    
    print("distance stats:", stats.describe(distances.flatten()))
    print()    
    
    ### compute travel times
    # set zero point for FMM zero contour
    speed = grid.copy()
    travel_times = calculate_travel_times(grid, speed, zero_point)
    
    plt.contour(travel_times, levels=50)
    plt.title(f'travel_times from {zero_point}')
    plt.colorbar()
    plt.show()
    
    plt.hist(travel_times.flatten(), bins=50)
    plt.title("travel_times")
    plt.show()
    
    print("travel_times stats:", stats.describe(travel_times.flatten()))
    print()
        
    
    ### Ganges image
    input_file_path = "data/BWGanges_BW.npy"
    grid = np.load(input_file_path)
    zero_point = (1400, 650)
    
    grid_masked = np.ma.masked_where(grid == 0, grid)
    
    distances = calculate_distances(grid_masked, zero_point)
    
    plt.contour(distances, levels=100)
    plt.colorbar()
    plt.title(f'distances from {zero_point}')
    plt.show()
    
    plt.hist(distances[distances > 0].flatten(), bins=100)
    plt.title("distances")
    plt.show()
    
    print("distance stats:", stats.describe(distances[distances > 0].flatten()))
    print()    
    
    travel_times = calculate_travel_times(grid_masked, grid_masked, zero_point)

    plt.contour(travel_times, levels=100)
    plt.title(f'travel_times from {zero_point}')
    plt.colorbar()
    plt.show()
    
    plt.hist(travel_times[travel_times > 0].flatten(), bins=50)
    plt.title("travel_times")
    plt.show()
    
    print(f"{input_file_path} travel_times stats:", stats.describe(travel_times[travel_times > 0].flatten()))
    print()
    
if __name__ == '__main__':
	main()


